@@ -27,13 +27,8 @@
 #include <gdk/gdkx.h>
 #include <gio/gio.h>
 
-#if GTK_CHECK_VERSION(3, 0, 0)
-#include <libdbusmenu-gtk3/menuitem.h>
-#include <libdbusmenu-gtk3/parser.h>
-#else
 #include <libdbusmenu-gtk/menuitem.h>
 #include <libdbusmenu-gtk/parser.h>
-#endif
 
 #include <libdbusmenu-glib/menuitem.h>
 #include <libdbusmenu-glib/server.h>
@@ -108,6 +103,35 @@ static GHashTable *rebuild_ids = NULL;
 static GDBusNodeInfo *      registrar_node_info = NULL;
 static GDBusInterfaceInfo * registrar_interface_info = NULL;
 
+#if HAVE_GTK3 == 0
+
+#include <X11/Xlib.h>
+
+/* This function is copied from GTK 3.4 and is not Copyright Canonical */
+static void
+gdk_x11_window_set_utf8_property  (GdkWindow *window,
+                                     const gchar *name,
+                                     const gchar *value)
+{
+	GdkDisplay *display;
+
+	display = gdk_window_get_display (window);
+
+	if (value != NULL) {
+		XChangeProperty (GDK_DISPLAY_XDISPLAY (display),
+		                 GDK_WINDOW_XID (window),
+		                 gdk_x11_get_xatom_by_name_for_display (display, name),
+		                 gdk_x11_get_xatom_by_name_for_display (display, "UTF8_STRING"), 8,
+		                 PropModeReplace, (guchar *)value, strlen (value));
+	} else {
+		XDeleteProperty (GDK_DISPLAY_XDISPLAY (display),
+		                 GDK_WINDOW_XID (window),
+		                 gdk_x11_get_xatom_by_name_for_display (display, name));
+	}
+}
+
+#endif
+
 static void
 activate_menu (AppMenuBridge *bridge,
                GtkWidget     *widget,
@@ -376,6 +400,15 @@ register_application_windows (AppMenuBridge *bridge)
                             context->cancel,
                             register_application_window_cb,
                             context);
+
+          GdkWindow * gdkwindow = gtk_widget_get_window(widget);
+          if (gdkwindow != NULL)
+            {
+              GDBusConnection * conn = g_dbus_proxy_get_connection(bridge->priv->appmenuproxy);
+
+              gdk_x11_window_set_utf8_property(gdkwindow, "_UBUNTU_APPMENU_UNIQUE_NAME", g_dbus_connection_get_unique_name(conn));
+              gdk_x11_window_set_utf8_property(gdkwindow, "_UBUNTU_APPMENU_OBJECT_PATH", context->path);
+            }
         }
     }
 }
@@ -519,9 +552,7 @@ rebuild (AppMenuBridge *bridge,
   data->bridge = bridge;
   data->widget = toplevel;
 
-  id = g_timeout_add (100,
-                      (GSourceFunc)do_rebuild,
-                      data);
+  id = gdk_threads_add_timeout (100, (GSourceFunc)do_rebuild, data);
 
   g_object_add_weak_pointer (G_OBJECT (data->widget), (gpointer*)&data->widget);
 
@@ -792,6 +823,8 @@ attach_notify_cb (GtkWidget     *widget,
       g_object_get (widget, "attach-widget", &attach, NULL);
 
       rebuild (bridge, attach);
+
+      g_clear_object (&attach);
     }
 }
 
@@ -852,9 +885,17 @@ app_menu_bridge_insert (UbuntuMenuProxy *proxy,
         {
           DbusmenuMenuitem *mi = dbusmenu_gtk_parse_get_cached_item (attach);
 
+          g_object_unref (attach);
+
           if (mi != NULL)
             {
               DbusmenuMenuitem *child_dmi = dbusmenu_gtk_parse_menu_structure (child);
+              DbusmenuMenuitem *child_parent = dbusmenu_menuitem_get_parent (child_dmi);
+
+              if (child_parent != NULL)
+                {
+                  dbusmenu_menuitem_child_delete (child_parent, child_dmi);
+                }
 
               g_object_set_data (G_OBJECT (child_dmi), "dbusmenu-parent", mi);
               dbusmenu_menuitem_child_add_position (mi,
@@ -912,7 +953,9 @@ app_menu_brige_shouldnt_load (void)
       || (g_strcmp0 (prg, "emacs") == 0)
       || (g_strcmp0 (prg, "emacs23") == 0)
       || (g_strcmp0 (prg, "emacs23-lucid") == 0)
-      || (g_strcmp0 (prg, "gnome-panel") == 0))
+      || (g_strcmp0 (prg, "glade") == 0)
+      || (g_strcmp0 (prg, "gnome-panel") == 0)
+      || (g_strcmp0 (prg, "Lotus Notes") == 0))
     {
       return TRUE;
     }
